import meshio
import pytest
from pathlib import Path

from ncollpyde import Volume

test_dir = Path(__file__).resolve().parent
project_dir = test_dir.parent
mesh_dir = project_dir / "meshes"


@pytest.fixture
def mesh():
    return meshio.read(str(mesh_dir / "teapot.stl"))


@pytest.fixture
def volume(mesh):
    return Volume.from_meshio(meshio, validate=True)


@pytest.fixture
def simple_mesh():
    return meshio.read(str(mesh_dir / "cube.stl"))


@pytest.fixture
def simple_volume(simple_mesh):
    return Volume.from_meshio(simple_mesh, validate=True)


@pytest.fixture
def sez_right():
    return Volume.from_meshio(
        meshio.read(str(mesh_dir / "SEZ_right.stl")), validate=True
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ncollpyde` package."""
from itertools import product

import pytest
import numpy as np
from trimesh import Trimesh

from ncollpyde import Volume


points_expected = [
    ([-2.3051376, -4.1556454, 1.9047838], True),  # internal
    ([-0.35222054, -0.513299, 7.6191354], False),  # external but in AABB
    ([0.13970017, 0.0928266, 4.7073355], True),  # vertex
    ([10, 10, 10], False),  # out of AABB
]


@pytest.mark.parametrize(["point", "expected"], points_expected)
def test_single(mesh, point, expected):
    """Sample pytest test function with the pytest fixture as an argument."""
    vol = Volume.from_meshio(mesh)
    assert (point in vol) == expected


def test_corner(mesh):
    vol = Volume.from_meshio(mesh)
    assert mesh.points[0] in vol


@pytest.mark.parametrize("threads", [None, 0, 2])
def test_many(mesh, threads):
    points = []
    expected = []
    for p, e in points_expected:
        points.append(p)
        expected.append(e)

    vol = Volume.from_meshio(mesh)
    assert np.array_equal(vol.contains(points, threads=threads), expected)


def test_no_validation(mesh):
    triangles = mesh.cells["triangle"]
    Volume(mesh.points, triangles, True)


def test_can_repair_hole(mesh):
    triangles = mesh.cells["triangle"]
    triangles = triangles[:-1]
    Volume(mesh.points, triangles, True)


def test_can_repair_inversion(mesh):
    triangles = mesh.cells["triangle"]
    triangles[-1] = triangles[-1, ::-1]
    Volume(mesh.points, triangles, True)


def test_can_repair_inversions(mesh):
    triangles = mesh.cells["triangle"]
    triangles = triangles[:, ::-1]
    Volume(mesh.points, triangles, True)


def test_points(mesh):
    points = mesh.points
    vol = Volume(mesh.points, mesh.cells["triangle"])

    actual = sorted(tuple(p) for p in vol.points)
    expected = sorted(tuple(p) for p in points)
    assert actual == expected


def test_triangles(mesh):
    points = mesh.points
    triangles = mesh.cells["triangle"]
    expected = Trimesh(points, triangles)

    vol = Volume(mesh.points, triangles)
    actual = Trimesh(vol.points, vol.faces)

    assert expected.volume == actual.volume


def test_extents(mesh):
    points = mesh.points
    vol = Volume(mesh.points, mesh.cells["triangle"])

    expected = np.array([points.min(axis=0), points.max(axis=0)], np.float32)
    actual = vol.extents

    assert np.allclose(expected, actual)


def test_points_roundtrip(mesh):
    points = mesh.points
    vol = Volume(mesh.points, mesh.cells["triangle"])

    expected = np.array(sorted(tuple(p) for p in points), dtype=np.float32)
    actual = np.array(sorted(tuple(p) for p in vol.points), dtype=np.float32)

    assert np.allclose(expected, actual)


def test_extents_validity(mesh):
    points = mesh.points
    vol = Volume(mesh.points, mesh.cells["triangle"])

    expected = np.array([points.min(axis=0), points.max(axis=0)], np.float32)
    actual = vol.extents
    assert np.allclose(expected, actual)


class ParamatrizationBuilder:
    def __init__(self, *names):
        self.param_names = names
        self.params = []
        self.test_names = []

    def add(self, test_name, *params):
        if len(params) != len(self.param_names):
            raise ValueError(
                f"Wrong number of params given"
                "(got {len(params)}, expected {len(self.param_names)})"
            )
        self.params.append(params)
        self.test_names.append(test_name)

    def as_dict(self):
        return {
            "argnames": self.param_names,
            "argvalues": self.params,
            "ids": self.test_names,
        }


params = ParamatrizationBuilder("coords", "is_internal")
middle = [0.5, 0.5, 0.5]
params.add("middle", middle, True)
for idx, corner in enumerate(product([0, 1], repeat=3)):
    params.add(f"corner-{idx}", corner, True)

for dim in range(3):
    this = middle.copy()
    this[dim] = 0
    params.add(f"face-low-dim{dim}", this.copy(), True)
    this[dim] = 1
    params.add(f"face-high-dim{dim}", this.copy(), True)

    this[dim] = -0.5
    params.add(f"external-low-dim{dim}", this.copy(), False)
    this[dim] = 1.5
    params.add(f"external-high-dim{dim}", this.copy(), False)


@pytest.mark.parametrize(**params.as_dict())
def test_cube(simple_volume, coords, is_internal):
    assert (coords in simple_volume) == is_internal


def test_issue3(sez_right):
    assert [28355.6, 51807.3, 47050] not in sez_right



from __future__ import annotations

import logging
import warnings
from numbers import Number
from typing import Union, Sequence, Optional, TYPE_CHECKING
from multiprocessing import cpu_count

import numpy as np

try:
    import trimesh
except ImportError:
    trimesh = None

from .ncollpyde import TriMeshWrapper, _precision

if TYPE_CHECKING:
    import meshio


logger = logging.getLogger(__name__)

N_CPUS = cpu_count()
DEFAULT_THREADS = 0

PRECISION = np.dtype(_precision())

ArrayLike1D = Union[np.ndarray, Sequence[Number]]
ArrayLike2D = Union[np.ndarray, Sequence[Sequence[Number]]]


class Volume:
    dtype = PRECISION
    threads = DEFAULT_THREADS

    def __init__(
        self,
        vertices: ArrayLike2D,
        triangles: ArrayLike2D,
        validate=False,
        threads=None,
    ):
        """
        Create a volume described by a triangular mesh with N vertices and M triangles.

        :param vertices: Nx3 array-like of floats, coordinates of triangle corners
        :param triangles: Mx3 array-like of ints,
            indices of ``vertices`` which describe each triangle
        :param validate: bool, whether to validate mesh.
            If trimesh is installed, the mesh is checked for watertightness and correct
            winding, and repairs made if possible.
            Otherwise, only very basic checks are made.
        :param validate: optional number or True, sets default threading for containment
            checks with this instance. Can also be set on the class.
        """
        vertices = np.asarray(vertices, self.dtype)
        triangles = np.asarray(triangles, np.uint64)
        if validate:
            vertices, triangles = self._validate(vertices, triangles)
        if threads is not None:
            self.threads = threads
        self._impl = TriMeshWrapper(vertices.tolist(), triangles.tolist())

    def _validate(self, vertices: np.ndarray, triangles: np.ndarray):
        if trimesh:
            tm = trimesh.Trimesh(vertices, triangles)
            if not tm.is_volume:
                logger.info("Mesh not valid, attempting to fix")
                tm.fill_holes()
                tm.fix_normals()
                if not tm.is_volume:
                    raise ValueError(
                        "Mesh is not a volume "
                        "(e.g. not watertight, incorrect winding) "
                        "and could not be fixed"
                    )

            return tm.vertices, tm.faces

        else:
            warnings.warn("trimesh not installed; full validation not possible")

            if vertices.shape[1:] != (3,):
                raise ValueError("Vertices are not in 3D")

            if triangles.shape[1:] != (3,):
                raise ValueError("Triangles do not have 3 points")

            if triangles.max() >= len(vertices):
                raise ValueError("Some triangle vertices do not exist in points")

            return vertices, triangles

    def __contains__(self, item: ArrayLike1D) -> bool:
        """Check whether a single point is in the volume.

        :param item:
        :return:
        """
        item = np.asarray(item, self.dtype)
        if item.shape != (3,):
            raise ValueError("Item is not a 3-length array-like")
        return self._impl.contains(item.tolist())

    def contains(
        self, coords: ArrayLike2D, threads: Optional[Union[int, bool]] = None
    ) -> np.ndarray:
        """Check whether multiple points (as a Px3 array-like) are in the volume.

        :param coords:
        :param threads: None,
            If ``threads`` is ``None``, the instance's ``threads`` attribute (default 0)
            is used.
            If ``threads`` is ``True``, ``threads`` is set to the number of CPUs.
            If ``threads`` is 0, the query will be done in serial (but the GIL will be
            released)
            If ``threads`` is something else (a number), the query will be parallelised
            over that number of threads.
        :return: np.ndarray of bools, whether each point is inside the volume
        """
        coords = np.asarray(coords, self.dtype)
        if coords.shape[1:] != (3,):
            raise ValueError("Coords is not a Nx3 array-like")

        if threads is None:
            threads = self.threads

        if not threads:
            out = self._impl.contains_many(coords.tolist())
        else:
            if threads is True:
                threads = N_CPUS
            out = self._impl.contains_many_threaded(coords.tolist(), threads)
        return np.array(out, dtype=bool)

    @classmethod
    def from_meshio(cls, mesh: "meshio.Mesh", validate=False, threads=None) -> Volume:
        """
        Convenience function for instantiating a Volume from a meshio Mesh.

        :param mesh: meshio Mesh whose only cells are triangles.
        :param validate: as passed to __init__, defaults to False
        :param threads: as passed to __init__, defaults to None
        :raises ValueError: if Mesh does not have triangle cells
        :return: Volume instance
        """
        try:
            return cls(mesh.points, mesh.cells["triangle"], validate, threads)
        except KeyError:
            raise ValueError("Must have triangle cells")

    @property
    def points(self) -> np.array:
        """
        Nx3 array of float32 describing vertices
        """
        return np.array(self._impl.points(), self.dtype)

    @property
    def faces(self) -> np.array:
        """
        Mx3 array of uint64 describing indexes into points array making up triangles
        """
        return np.array(self._impl.faces(), np.uint64)

    @property
    def extents(self) -> np.array:
        """
        [
            [xmin, ymin, zmin],
            [xmax, ymax, zmax],
        ]
        """
        return np.array(self._impl.aabb(), self.dtype)



# -*- coding: utf-8 -*-

"""Top-level package for ncollpyde."""

__author__ = """Chris L. Barnes"""
__email__ = "chrislloydbarnes@gmail.com"

from .ncollpyde import _version
from .main import Volume

__version__ = _version()
__version_info__ = tuple(int(n) for n in __version__.split("-")[0].split("."))

__all__ = ["Volume"]



from pathlib import Path

import numpy as np
import meshio
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from ncollpyde import Volume


MESH_PATH = Path("meshes/teapot.stl")

vol = Volume.from_meshio(meshio.read(str(MESH_PATH)), validate=True, threads=True)

extents = vol.extents
size = np.diff(extents, axis=0).flatten()
extents[0] -= size * 0.1
extents[1] += size * 0.1
new_size = np.diff(extents, axis=0).flatten()

rng = np.random.RandomState(1991)
samples = rng.random_sample((1_000_000, 3)) * new_size + extents[0]

print("containment check")
idxs = vol.contains(samples, True)
print("containment check done")
internals = samples[idxs]
externals = samples[~idxs]

fig: Figure = plt.figure()
ax: Axes3D = fig.add_subplot(111, projection="3d")

ax.set_xlim(extents[:, 0])
ax.set_ylim(extents[:, 1])
ax.set_zlim(extents[:, 2])
ax.scatter(internals[:, 0], internals[:, 1], internals[:, 2])

plt.show()



import numpy as np
import trimesh

import meshio
import pytest

from ncollpyde import Volume

SAMPLES_PER_DIM = 10
PADDING = 0.2
THREADS = 3
ITERATIONS = 20


class PyOctreeWrapper:
    """Wrapper around pyoctree's ray casting for volume checks.

    Adapted from https://github.com/schlegelp/pymaid/blob/master/pymaid/intersect.py
    """

    def __init__(self, vertices, faces, octree):
        self.vertices = np.asarray(vertices, dtype=np.float32, order="C")
        self.faces = np.asarray(faces, dtype=np.int32, order="C")

        # pass in so that the class does not need to re-import pyoctree
        self.octree = octree

    def contains(self, points, multi_ray=False):
        """Use pyoctree's raycsasting to test if points are within volume."""

        # Remove points outside of bounding box
        isin = self._in_bbox(points)
        in_points = points[isin]

        # mx = np.array(volume.vertices).max(axis=0)
        mn = np.array(self.vertices).min(axis=0)

        # Perform ray intersection on points inside bounding box
        rayPointList = np.array(
            [[[p[0], mn[1], mn[2]], p] for p in in_points], dtype=np.float32
        )

        # Get intersections and extract coordinates of intersection
        intersections = [
            np.array([i.p for i in self.octree.rayIntersection(ray)])
            for ray in rayPointList
        ]

        # In a few odd cases we can get the multiple intersections at the exact
        # same coordinate (something funny with the faces).
        unique_int = [
            np.unique(np.round(i), axis=0) if np.any(i) else i for i in intersections
        ]

        # Unfortunately rays are bidirectional -> we have to filter intersections
        # to those that occur "above" the point we are querying
        unilat_int = [
            i[i[:, 2] >= p] if np.any(i) else i
            for i, p in zip(unique_int, in_points[:, 2])
        ]

        # Count intersections
        int_count = [i.shape[0] for i in unilat_int]

        # Get odd (= in volume) numbers of intersections
        is_odd = np.remainder(int_count, 2) != 0

        # If we want to play it safe, run the above again with two additional rays
        # and find a majority decision.
        if multi_ray:
            # Run ray from left back
            rayPointList = np.array(
                [[[mn[0], p[1], mn[2]], p] for p in in_points], dtype=np.float32
            )
            intersections = [
                np.array([i.p for i in self.octree.rayIntersection(ray)])
                for ray in rayPointList
            ]
            unique_int = [
                np.unique(i, axis=0) if np.any(i) else i for i in intersections
            ]
            unilat_int = [
                i[i[:, 0] >= p] if np.any(i) else i
                for i, p in zip(unique_int, in_points[:, 0])
            ]
            int_count = [i.shape[0] for i in unilat_int]
            is_odd2 = np.remainder(int_count, 2) != 0

            # Run ray from lower left
            rayPointList = np.array(
                [[[mn[0], mn[1], p[2]], p] for p in in_points], dtype=np.float32
            )
            intersections = [
                np.array([i.p for i in self.octree.rayIntersection(ray)])
                for ray in rayPointList
            ]
            unique_int = [
                np.unique(i, axis=0) if np.any(i) else i for i in intersections
            ]
            unilat_int = [
                i[i[:, 1] >= p] if np.any(i) else i
                for i, p in zip(unique_int, in_points[:, 1])
            ]
            int_count = [i.shape[0] for i in unilat_int]
            is_odd3 = np.remainder(int_count, 2) != 0

            # Find majority consensus
            is_odd = is_odd.astype(int) + is_odd2.astype(int) + is_odd3.astype(int)
            is_odd = is_odd >= 2

        isin[isin] = is_odd
        return isin

    def _in_bbox(self, points):
        """Test if points are within a bounding box.
        Parameters
        ----------
        points :    numpy array
                    2D  array of xyz coordinates.
        bbox :      numpy array | pymaid.Volume
                    If Volume will use the volumes bounding box. If numpy array
                    must be either::
                        [[x_min, x_max], [y_min, y_max], [z_min, z_max]]
                        [[x_min, y_min, z_min], [x_max, y_max, z_max]]
        Returns
        -------
        numpy array
                    Boolean array with True/False for each point.
        """
        mx = np.array(self.vertices).max(axis=0)
        mn = np.array(self.vertices).min(axis=0)

        # Get points outside of bounding box
        bbox_out = (points >= mx).any(axis=1) | (points <= mn).any(axis=1)
        isin = ~bbox_out

        return isin


@pytest.fixture
def pyoctree_volume(mesh):
    try:
        from pyoctree import pyoctree
    except ImportError:
        pytest.skip("pyoctree not available")
    else:
        vertices = np.asarray(mesh.points, dtype=np.float64, order="C")
        faces = np.asarray(mesh.cells["triangle"], dtype=np.int32, order="C")
        octree = pyoctree.PyOctree(vertices, faces)
        return PyOctreeWrapper(vertices, faces, octree)


@pytest.fixture
def ncollpyde_volume(mesh):
    return Volume.from_meshio(mesh)


@pytest.fixture
def trimesh_volume(mesh):
    try:
        import trimesh
    except ImportError:
        pytest.skip("trimesh not available")
    else:
        return trimesh.Trimesh(vertices=mesh.points, faces=mesh.cells["triangle"])


@pytest.fixture
def expected(trimesh_volume, sample_points):
    return trimesh_volume.contains(sample_points)


def make_sample_points(
    aabb: np.ndarray, padding=PADDING, samples_per_dim=SAMPLES_PER_DIM
):
    """[[minx, miny, minz], [max, maxy, maxz]]"""
    mins, maxes = aabb
    ranges = maxes - mins
    mins -= ranges * padding
    maxes += ranges * padding
    sample1d = np.linspace(mins, maxes, samples_per_dim, dtype=Volume.dtype, axis=1)
    grid = np.meshgrid(*sample1d)
    return np.array([arr.flatten() for arr in grid]).T


@pytest.fixture
def sample_points(mesh: meshio.Mesh):
    return make_sample_points([mesh.points.min(axis=0), mesh.points.max(axis=0)])


def check_internals_equal(expected, actual):
    assert expected.sum() == actual.sum()
    assert np.array_equal(expected, actual)


def test_trimesh(trimesh_volume: trimesh.Trimesh, sample_points, expected, benchmark):
    actual = benchmark(trimesh_volume.contains, sample_points)
    check_internals_equal(expected, actual)


def test_ncollpyde(ncollpyde_volume: Volume, sample_points, expected, benchmark):
    actual = benchmark(ncollpyde_volume.contains, sample_points)
    check_internals_equal(expected, actual)


# def test_ncollpyde_threaded(
#     ncollpyde_volume: Volume, sample_points, expected, benchmark
# ):
#     actual = benchmark(ncollpyde_volume.contains, sample_points, threads=THREADS)
#     check_internals_equal(expected, actual)


# def test_ncollpyde_complex(sez_right: Volume, benchmark):
#     sample_points = make_sample_points(sez_right.extents, 0.1, 100)
#     benchmark(sez_right.contains, sample_points)


# def test_ncollpyde_complex_threads(sez_right: Volume, benchmark):
#     sample_points = make_sample_points(sez_right.extents, 0.1, 100)
#     benchmark(sez_right.contains, sample_points, threads=THREADS)


@pytest.mark.xfail(reason="Results do not match yet")
def test_pyoctree(pyoctree_volume: PyOctreeWrapper, sample_points, expected, benchmark):
    actual = benchmark(pyoctree_volume.contains, sample_points)
    check_internals_equal(expected, actual)
